@@ -7,17 +7,20 @@ import json
 import shutil
 import subprocess
 
+from ladybug.futil import preparedir, write_to_file
+from ladybug.epw import EPW
+from honeybee_energy.config import folders as hb_energy_folders
+from honeybee_energy.result.emissions import emissions_region
+
 from .config import folders
 from .measure import MapperMeasure
 from .reopt import REoptParameter
 
-from honeybee_energy.config import folders as hb_energy_folders
-from ladybug.futil import preparedir, write_to_file
-
 
 def base_honeybee_osw(
         project_directory, sim_par_json=None, additional_measures=None,
-        additional_mapper_measures=None, base_osw=None, epw_file=None, skip_report=True):
+        additional_mapper_measures=None, base_osw=None, epw_file=None,
+        skip_report=True, emissions_year=None):
     """Create a honeybee_workflow.osw to be used as a base in URBANopt simulations.
 
     This method will also copy the Honeybee.rb mapper to this folder if it is
@@ -48,7 +51,11 @@ def base_honeybee_osw(
             measure skipped as part of the workflow. If False, the measure will
             be run after all simulations are complete. Note that this input
             has no effect if the default_feature_reports measure is already
-            in the base_osw or additional_measures (Default: True)
+            in the base_osw or additional_measures. (Default: True).
+        emissions_year: An optional integer to set the year for which carbon emissions
+            will be computed. If not for a historical year, values must be an even
+            number and be between 2020 and 2050. If None, no carbon emission
+            calculations will be included in the simulation. (Default: None).
 
     Returns:
         The file path to the honeybee_workflow.osw written out by this method.
@@ -89,6 +96,28 @@ def base_honeybee_osw(
         m_dir = os.path.join(hb_energy_folders.honeybee_openstudio_gem_path, 'measures')
         osw_dict['measure_paths'].append(m_dir)
 
+    # add the emissions reporting if a year has been selected
+    if emissions_year is not None and epw_file is not None:
+        epw_obj = EPW(epw_file)
+        ems_region = emissions_region(epw_obj.location)
+        if ems_region is not None:
+            if emissions_year < 2020:
+                hist_yr, fut_yr = emissions_year, 2030
+            else:
+                hist_yr, fut_yr = 2019, emissions_year
+            emissions_measure_dict = {
+                'arguments': {
+                    'future_subregion': ems_region[0],
+                    'hourly_historical_subregion': ems_region[2],
+                    'annual_historical_subregion': ems_region[1],
+                    'future_year': str(fut_yr),
+                    'hourly_historical_year': '2019',
+                    'annual_historical_year': str(hist_yr)
+                },
+                'measure_dir_name': 'add_ems_emissions_reporting'
+            }
+            osw_dict['steps'].append(emissions_measure_dict)
+
     # add any additional measures to the osw_dict
     if additional_measures or additional_mapper_measures:
         measures = []
@@ -112,7 +141,7 @@ def base_honeybee_osw(
         for m_path in measure_paths:  # add outside measure paths
             osw_dict['measure_paths'].append(m_path)
 
-    # add default feature reports if they aren't already in the steps
+    # add default feature reports if they aren't in the steps
     all_measures = [step['measure_dir_name'] for step in osw_dict['steps']]
     if 'default_feature_reports' not in all_measures:
         report_measure_dict = {
@@ -197,12 +226,15 @@ def prepare_urbanopt_folder(feature_geojson, cpu_count=None, verbose=False):
     return _make_scenario(feature_geojson)
 
 
-def run_urbanopt(feature_geojson, scenario_csv):
+def run_urbanopt(feature_geojson, scenario_csv, cpu_count=None):
     """Run a feature and scenario file through URBANopt on any operating system.
 
     Args:
         feature_geojson: The full path to a .geojson file containing the
             footprints of buildings to be simulated.
+        cpu_count: A positive integer for the number of CPUs to use in the
+            simulation. If None, the simulation will default to whatever is
+            specified in the runner.conf. (Default: None).
         scenario_csv: The full path to  a .csv file for the URBANopt scenario.
 
     Returns:
@@ -229,9 +261,9 @@ def run_urbanopt(feature_geojson, scenario_csv):
     folders.check_urbanopt_version()
     # run the simulation
     if os.name == 'nt':  # we are on Windows
-        directory = _run_urbanopt_windows(feature_geojson, scenario_csv)
+        directory = _run_urbanopt_windows(feature_geojson, scenario_csv, cpu_count)
     else:  # we are on Mac, Linux, or some other unix-based system
-        directory = _run_urbanopt_unix(feature_geojson, scenario_csv)
+        directory = _run_urbanopt_unix(feature_geojson, scenario_csv, cpu_count)
 
     # output the simulation files
     return _output_urbanopt_files(directory)
@@ -309,7 +341,7 @@ def run_reopt(feature_geojson, scenario_csv, urdb_label, reopt_parameters=None,
     project_folder = os.path.dirname(feature_geojson)
 
     # write the parameter file
-    if reopt_parameters is None:  # genrate some defaults
+    if reopt_parameters is None:  # generate some defaults
         reopt_parameters = REoptParameter()
         reopt_parameters.pv_parameter.max_kw = 1000000000
         reopt_parameters.storage_parameter.max_kw = 1000000
@@ -336,6 +368,62 @@ def run_reopt(feature_geojson, scenario_csv, urdb_label, reopt_parameters=None,
     return _output_reopt_files(directory)
 
 
+def run_rnm(feature_geojson, scenario_csv, underground_ratio=0.9, lv_only=True,
+            nodes_per_building=1):
+    """Run a feature and scenario file through RNM post processing.
+
+    Note that the URBANopt simulation must already be run with the input feature_geojson
+    and scenario_csv in order for the RNM post-processing to be successful.
+
+    Args:
+        feature_geojson: The full path to a .geojson file containing the
+            footprints of buildings to be simulated.
+        scenario_csv: The full path to a .csv file for the URBANopt scenario.
+        underground_ratio: A number between 0 and 1 for the ratio of overall cables
+            that are underground vs. overhead in the analysis. (Default: 0.9).
+        lv_only: Boolean to note whether to consider only low voltage consumers
+            in the analysis. (Default: True).
+        nodes_per_building: Positive integer for the maximum number of low voltage
+            nodes to represent a single building. (Default: 1).
+
+    Returns:
+        Path to a folder that contains all of the RNM output files.
+    """
+    # load the information from the GeoJSON
+    geo_dict, project_dict = None, None
+    if os.path.isfile(feature_geojson):
+        with open(feature_geojson, 'r') as fg:
+            geo_dict = json.load(fg)
+            project_dict = geo_dict['project']
+    # change the GeoJSON to have the RNM inputs
+    if underground_ratio != 0.9 or not lv_only or nodes_per_building != 1:
+        if geo_dict is not None:
+            geo_dict['project']['underground_cables_ratio'] = underground_ratio
+            geo_dict['project']['only_lv_consumers'] = lv_only
+            geo_dict['project']['max_number_of_lv_nodes_per_building'] = \
+                nodes_per_building
+            with open(feature_geojson, 'w') as fp:
+                json.dump(geo_dict, fp, indent=4)
+    # run the simulation
+    folders.check_urbanopt_version()
+    if os.name == 'nt':  # we are on Windows
+        directory = _run_rnm_windows(feature_geojson, scenario_csv)
+    else:  # we are on Mac, Linux, or some other unix-based system
+        directory = _run_rnm_unix(feature_geojson, scenario_csv)
+    # get the path to the results folder
+    scenario_name = os.path.basename(scenario_csv).replace('.csv', '')
+    rnm_path = os.path.join(directory, 'run', scenario_name, 'rnm-us', 'results')
+    # copy the project information into the RNM GeoJSON
+    if os.path.isdir(rnm_path):
+        rnm_geojson = os.path.join(rnm_path, 'GeoJSON', 'Distribution_system.json')
+        with open(rnm_geojson, 'r') as fg:
+            rnm_dict = json.load(fg)
+        rnm_dict['project'] = project_dict
+        with open(rnm_geojson, 'w') as fp:
+            json.dump(rnm_dict, fp, indent=4)
+        return rnm_path
+
+
 def _add_mapper_measure(project_directory, mapper_measure):
     """Add mapper measure arguments to a geoJSON and the mapper_measures.json.
 
@@ -438,7 +526,7 @@ def _recommended_processor_count():
     return 1 if cpu_count is None or cpu_count <= 1 else cpu_count - 1
 
 
-def _run_urbanopt_windows(feature_geojson, scenario_csv):
+def _run_urbanopt_windows(feature_geojson, scenario_csv, cpu_count):
     """Run a feature and scenario file through URBANopt on a Windows-based os.
 
     A batch file will be used to run the simulation.
@@ -447,6 +535,7 @@ def _run_urbanopt_windows(feature_geojson, scenario_csv):
         feature_geojson: The full path to a .geojson file containing the
             footprints of buildings to be simulated.
         scenario_csv: The full path to  a .csv file for the URBANopt scenario.
+        cpu_count: A positive integer for the number of processors to use.
 
     Returns:
         Path to the folder out of which the simulation was run.
@@ -455,17 +544,17 @@ def _run_urbanopt_windows(feature_geojson, scenario_csv):
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the batch file to call URBANopt CLI
     working_drive = directory[:2]
-    batch = '{}\ncd {}\ncall {}\nuo run -f {} -s {}'.format(
+    batch = '{}\ncd {}\ncall "{}"\nuo run -f "{}" -s "{}"'.format(
         working_drive, working_drive, folders.urbanopt_env_path,
         feature_geojson, scenario_csv)
     batch_file = os.path.join(directory, 'run_simulation.bat')
     write_to_file(batch_file, batch, True)
     # run the batch file
-    os.system(batch_file)
+    os.system('"{}"'.format(batch_file))
     return directory
 
 
-def _run_urbanopt_unix(feature_geojson, scenario_csv):
+def _run_urbanopt_unix(feature_geojson, scenario_csv, cpu_count):
     """Run a feature and scenario file through URBANopt on a Unix-based os.
 
     This includes both Mac OS and Linux since a shell will be used to run
@@ -475,6 +564,7 @@ def _run_urbanopt_unix(feature_geojson, scenario_csv):
         feature_geojson: The full path to a .geojson file containing the
             footprints of buildings to be simulated.
         scenario_csv: The full path to  a .csv file for the URBANopt scenario.
+        cpu_count: A positive integer for the number of processors to use.
 
     Returns:
         Path to the folder out of which the simulation was run.
@@ -482,7 +572,7 @@ def _run_urbanopt_unix(feature_geojson, scenario_csv):
     # check the input file
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the shell script to call URBANopt CLI
-    shell = '#!/usr/bin/env bash\nsource {}\nuo run -f {} -s {}'.format(
+    shell = '#!/usr/bin/env bash\nsource "{}"\nuo run -f "{}" -s "{}"'.format(
         folders.urbanopt_env_path, feature_geojson, scenario_csv)
     shell_file = os.path.join(directory, 'run_simulation.sh')
     write_to_file(shell_file, shell, True)
@@ -611,13 +701,13 @@ def _run_default_report_windows(feature_geojson, scenario_csv):
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the batch file to call URBANopt CLI
     working_drive = directory[:2]
-    batch = '{}\ncd {}\ncall {}\nuo process --default -f {} -s {}'.format(
+    batch = '{}\ncd {}\ncall "{}"\nuo process --default -f "{}" -s "{}"'.format(
         working_drive, working_drive, folders.urbanopt_env_path,
         feature_geojson, scenario_csv)
     batch_file = os.path.join(directory, 'run_default_report.bat')
     write_to_file(batch_file, batch, True)
     # run the batch file and return output files
-    os.system(batch_file)
+    os.system('"{}"'.format(batch_file))
     result_folder = os.path.basename(scenario_csv).lower().replace('.csv', '')
     run_folder = os.path.join(directory, 'run', result_folder)
     return os.path.join(run_folder, 'default_scenario_report.csv'), \
@@ -638,8 +728,8 @@ def _run_default_report_unix(feature_geojson, scenario_csv):
     # check the input file
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the shell script to call OpenStudio CLI
-    shell = '#!/usr/bin/env bash\nsource {}\n' \
-        'uo process --default -f {} -s {}'.format(
+    shell = '#!/usr/bin/env bash\nsource "{}"\n' \
+        'uo process --default -f "{}" -s "{}"'.format(
             folders.urbanopt_env_path, feature_geojson, scenario_csv)
     shell_file = os.path.join(directory, 'run_default_report.sh')
     write_to_file(shell_file, shell, True)
@@ -672,14 +762,14 @@ def _run_reopt_windows(feature_geojson, scenario_csv, developer_key):
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the batch file to call URBANopt CLI
     working_drive = directory[:2]
-    batch = '{}\ncd {}\ncall {}\nSET GEM_DEVELOPER_KEY={}\n' \
-        'uo process --reopt-scenario -f {} -s {}'.format(
+    batch = '{}\ncd {}\ncall "{}"\nSET GEM_DEVELOPER_KEY={}\n' \
+        'uo process --reopt-scenario -f "{}" -s "{}"'.format(
             working_drive, working_drive, folders.urbanopt_env_path, developer_key,
             feature_geojson, scenario_csv)
     batch_file = os.path.join(directory, 'run_reopt.bat')
     write_to_file(batch_file, batch, True)
     # run the batch file
-    os.system(batch_file)
+    os.system('"{}"'.format(batch_file))
     result_folder = os.path.basename(scenario_csv).lower().replace('.csv', '')
     return os.path.join(directory, 'run', result_folder)
 
@@ -702,8 +792,8 @@ def _run_reopt_unix(feature_geojson, scenario_csv, developer_key):
     # check the input file
     directory = _check_urbanopt_file(feature_geojson, scenario_csv)
     # Write the shell script to call OpenStudio CLI
-    shell = '#!/usr/bin/env bash\nsource {}\nGEM_DEVELOPER_KEY={}\n' \
-        'uo process --reopt-scenario -f {} -s {}'.format(
+    shell = '#!/usr/bin/env bash\nsource "{}"\nGEM_DEVELOPER_KEY={}\n' \
+        'uo process --reopt-scenario -f "{}" -s "{}"'.format(
             folders.urbanopt_env_path, developer_key, feature_geojson, scenario_csv)
     shell_file = os.path.join(directory, 'run_reopt.sh')
     write_to_file(shell_file, shell, True)
@@ -736,3 +826,59 @@ def _output_reopt_files(directory):
     csv = csv_file if os.path.isfile(csv_file) else None
     report_json = report_json_file if os.path.isfile(report_json_file) else None
     return csv, report_json
+
+
+def _run_rnm_windows(feature_geojson, scenario_csv):
+    """Run a feature and scenario file through RNM on a Windows-based os.
+
+    A batch file will be used to run the simulation.
+
+    Args:
+        feature_geojson: The full path to a .geojson file containing the
+            footprints of buildings to be simulated.
+        scenario_csv: The full path to  a .csv file for the URBANopt scenario.
+
+    Returns:
+        Path to the project folder.
+    """
+    # check the input file
+    directory = _check_urbanopt_file(feature_geojson, scenario_csv)
+    # Write the batch file to call URBANopt CLI
+    working_drive = directory[:2]
+    batch = '{}\ncd {}\ncall "{}"\nuo rnm --feature "{}" --scenario "{}"'.format(
+        working_drive, working_drive, folders.urbanopt_env_path,
+        feature_geojson, scenario_csv)
+    batch_file = os.path.join(directory, 'run_rnm.bat')
+    write_to_file(batch_file, batch, True)
+    # run the batch file
+    os.system('"{}"'.format(batch_file))
+    return directory
+
+
+def _run_rnm_unix(feature_geojson, scenario_csv):
+    """Run a feature and scenario file through RNM on a Unix-based os.
+
+    This includes both Mac OS and Linux since a shell script will be used to run
+    the simulation.
+
+    Args:
+        feature_geojson: The full path to a .geojson file containing the
+            footprints of buildings to be simulated.
+        scenario_csv: The full path to  a .csv file for the URBANopt scenario.
+
+    Returns:
+        Path to the project folder.
+    """
+    # check the input file
+    directory = _check_urbanopt_file(feature_geojson, scenario_csv)
+    # Write the shell script to call URBANopt CLI
+    shell = '#!/usr/bin/env bash\nsource "{}"\nuo rnm --feature "{}" -s-scenario ' \
+        '"{}"'.format(folders.urbanopt_env_path, feature_geojson, scenario_csv)
+    shell_file = os.path.join(directory, 'run_rnm.sh')
+    write_to_file(shell_file, shell, True)
+    # make the shell script executable using subprocess.check_call
+    # this is more reliable than native Python chmod on Mac
+    subprocess.check_call(['chmod', 'u+x', shell_file])
+    # run the shell script
+    subprocess.call(shell_file)
+    return directory
